from ..service import ReceivingService, ThreadService
from ..wire import RawWire

import wave

class SaveWav(ThreadService, ReceivingService):
    def __init__(self, filename=None):
        super(SaveWav, self).__init__()

        self.input_wire=RawWire()
        
        if filename is None:
            filename="data.wav"
        
        self.outputfile=filename
        
        self.output=wave.open(self.outputfile, 'wb')

    def start(self):
        self.output.setnchannels(1)
        if self.input_wire.FORMAT!="int16":
            raise Exception("Format to be saved in wav needs to be int16")
        self.output.setsampwidth(2)
        self.output.setframerate(self.input_wire.RATE)

        ThreadService.start(self)



    def process(self, data):
        self.output.writeframes(data)

    def stop(self):
        self.output.close()
        ThreadService.stop(self)



import numpy
from ..service import ReceivingService, ThreadService
from ..wire import RawWire

try:
    import soundcard
    HAS_SOUNDCARD=True
except ImportError:
    HAS_SOUNDCARD=False

class SoundCardPlay(ThreadService, ReceivingService):
    def __init__(self, frames_per_buffer=2048, output_device_index=None, output_device_name=""):
        if not HAS_SOUNDCARD:
          raise Exception("soundcard module not or not correctly installed")        
        self.input_wire=RawWire()
        self.frames_per_buffer=frames_per_buffer
        self.output_device_index=output_device_index
        self.output_device_name=output_device_name
        super(SoundCardPlay, self).__init__()

    def connect_input(self,service):
        super(SoundCardPlay, self).connect_input(service)
        if self.input_wire.FORMAT != "float32":
            raise Exception("SoundCard playback only supports float32 format")

    def _process(self):
        speaker=soundcard.get_speaker(self.output_device_index or self.output_device_name)
        print(speaker)
        with speaker.player(samplerate=self.input_wire.RATE, blocksize=self.frames_per_buffer) as player:
            while not self.done:
                data=self.receive_input()
                if data is None:
                    self.done=True
                else:
                    data=numpy.frombuffer(data,dtype=self.input_wire.FORMAT)

                if not self.stopped and not self.done:
                    try:
                        player.play(data)
                    except Exception as ex:
                        self.print_message( "error: {0}".format(str(ex)))
                        self.done=True
        self.stopped=True
        



from ..service import ReceivingService, SourceService, ThreadService
from ..wire import FloatWire, PulseWire

from functools import partial
import numpy

try:
    import scipy
    from scipy.optimize import curve_fit
    HAS_SCIPY=True    
except:
    HAS_SCIPY=False
    
class PulseDetection(ThreadService, SourceService, ReceivingService):
    def __init__(self, threshold=0.005, window=24*1024, debug=False):
        super(PulseDetection, self).__init__()
        self.input_wire=FloatWire()
        self.threshold=threshold
        self.window=window
        self.debug=debug
        self._x=numpy.arange(window)

    def output_protocol(self, wire):
        assert isinstance(wire, PulseWire)
        wire._debug=self.debug

    def start(self):
        self.data=numpy.zeros(self.window, dtype=self.input_wire.FORMAT)
        self.ndata=0
        self.RATE=self.input_wire.RATE
        self.itime=0
        ThreadService.start(self)
  
    def amplitude_and_quality(self, start,end):
        return numpy.max(self.data[start:end]),0

    def detect_pulses(self):
        larger=self.data>=self.threshold
        smaller=self.data<self.threshold
        
        up=smaller[:-1]*larger[1:]
        down=larger[:-1]*smaller[1:]
        
        up=numpy.argwhere(up) 
        down=numpy.argwhere(down) 
        
        up=up[:,0]
        down=down[:,0]
                
        if larger[0]: # discard first down crossing edge
          down=down[1:]
        if larger[-1]: # discard last up crossing edge
          up=up[:-1]
        
        if len(up)!=len(down):
            raise Exception("at this point, number of up and down edges must be same")
       
        #~ assert numpy.all(up<down)
        #~ assert numpy.all(up[1:]>down[:-1])
        #~ print "number of pulses:", len(up)
        
        pulses=[]
        for start,end in zip(up,down):
          if end-start>5:
            time=start/(1.*self.RATE)+self.itime
            try:
              amplitude, sigma=self.amplitude_and_quality(start,end)
            except:
              self.print_message("detection error")
              raise
              continue
            
            width=end-start

            if self.debug:
                pulse=self.data[max(start-5,0):end+15].copy()
                pulses.append((time,amplitude, width, sigma, pulse))
            else:
                pulses.append((time,amplitude, width, sigma))

        self.itime+=self.window/(1.*self.RATE)
        return pulses

    def process(self, data):
        outdata=[]
        while len(data)>0:
            start=self.ndata
            end=min(self.window,start+len(data))
            self.data[start:end]=data[:end-start]
            data=data[end-start:]
            self.ndata=end
            if self.ndata>=self.window:
                outdata.extend(self.detect_pulses())
                self.ndata=0
        
        return outdata or None


def pulse(x, A, x0,  y, tau):
        return y+(x>=x0)*A*(x-x0)**2*numpy.exp(-(x/tau))+(x<x0)*0.


class FittedPulseDetection(PulseDetection):
    """ 
    detects pulses and determines amplitude by fitting a predifined 
    pulse shape. x**2*exp(-x/tau).

    In principle there could be two reasons to use fitted pulse shapes: in 
    case the signal is undersampled the maximum can fall between two 
    samples or if the signal saturates the ADC and the maximum can still be 
    recovered by fitting the rising and falling edges. 

    """
    def __init__(self, threshold=0.005, window=24*1024, debug=False, 
                  fit_threshold=0., signal_noise=0.0035, pulse_decay_time=2.7247/48000):
        if not HAS_SCIPY:
            raise Exception("pulse fitting needs Scipy...")
        super(FittedPulseDetection, self).__init__(threshold=threshold, 
                                                   window=window, debug=debug)

        self.print_message("currently assumes very specific pulse shape, x**2 exp(-x/tau)")

        self.fit_threshold=fit_threshold
        self.sigma=signal_noise
        self.pulse_decay_time=pulse_decay_time

    def _fit_pulse(self, x, yp):

        A_=numpy.max(yp) # initial amplitude
        x0_=4.5 # initial timeshift
        tau=self.pulse_decay_time*self.input_wire.RATE
        
        template=partial(pulse, y=0., tau=tau) # fix base level (=0) and decay time
        
        a=(yp<0.95) * (yp>A_/5.) # select range
        x_=x[a]
        yp_=yp[a]
        
        sigma_=self.sigma*numpy.ones_like(x_) # weights
      
        (A,x0),cov=curve_fit(template, x_,yp_, p0=(A_,x0_), sigma=sigma_)
          
        AA=A*numpy.exp(-x0/tau)
        e=4*AA*tau**2*numpy.exp(-2.) # peak value
    
        p_=template(x_,A,x0)
        sigma_=numpy.sum((p_-yp_)**2)/len(yp_)
        sigma_=sigma_**0.5 / self.sigma # normalized mean error
    
        return e, sigma_
  
    def amplitude_and_quality(self, start,end):
        data=self.data[max(start-5,0):end+15]
        max_data=numpy.max(self.data[start:end])
        if max_data>self.fit_threshold:
            return self._fit_pulse(self._x[:len(data)], data)
        else:
            return max_data,0.



from ..service import ThreadService, SourceService
from ..wire import RawWire

import time
import numpy
import wave

pyaudio_nbytes=dict(int16=2, float32=4)

class RawReplay(ThreadService, SourceService):
    def __init__(self, filename="data.raw", frames_per_buffer=1024, realtime=True, 
        sample_rate=48000, sample_format="float32"):
        super(RawReplay, self).__init__()
        self.CHANNELS=1
        self.RATE=sample_rate
        self.FORMAT=sample_format
        self.filename=filename
        self.frames_per_buffer=frames_per_buffer
        self.realtime=realtime

    def output_protocol(self, wire):
        assert isinstance(wire, RawWire)
        wire.CHANNELS=self.CHANNELS
        wire.RATE=self.RATE
        wire.FORMAT=self.FORMAT
    
    def readframes(self,n):
        return self._file.read(pyaudio_nbytes[self.FORMAT]*n)
    
    def process(self, data=None):
        if self.realtime:
          time.sleep(self.frames_per_buffer/(1.*self.RATE))
          t=time.time()
          dt=t-self.t0
          self.t0=t
          n=int(dt*self.RATE/16)*16
        else:
          n=int(self.frames_per_buffer/16)*16
        data=self.readframes(n)
        if len(data)==0:
          data=None
          self.done=True
        return data
        
    def start(self):
        self.t0=time.time()
        self._file=open(self.filename, "rb")
        ThreadService.start(self)

    def close(self):
        ThreadService.close(self)
        self._file.close()


format_from_width={2 : "int16", 4 : "float32"}

class WavReplay(RawReplay):
    def __init__(self, filename, frames_per_buffer=1024, realtime=True):
        super(WavReplay, self).__init__()
        self.filename=filename
        self.frames_per_buffer=frames_per_buffer
        self.realtime=realtime
        self._file=wave.open(self.filename, "rb")
        self.CHANNELS=self._file.getnchannels()==1
        self.RATE=self._file.getframerate()
        self.SAMPLEWIDTH=self._file.getsampwidth()
        self.FORMAT=format_from_width[self.SAMPLEWIDTH]

    def readframes(self,n):
        return self._file.readframes(n)
        
    def start(self):
        self.t0=time.time()
        ThreadService.start(self)

def FileReplay(filename, **kwargs):
    if filename.endswith("wav"):
        return WavReplay(filename, **kwargs)
    return RawReplay(filename, **kwargs)



#!/usr/bin/env python
"""
gammalab-histogram makes a histogram of detected peak heights. 
"""

import argparse

try:
  import matplotlib
  matplotlib.use('TkAgg')
except:
  pass

from gammalab import main
from gammalab.acquisition import PyAudio, SoundCard, FileReplay
from gammalab.transform import Raw2Float, Scale
from gammalab.analysis import PulseDetection, FittedPulseDetection
from gammalab.analysis import AggregateHistogram
from gammalab.analysis import Count
from gammalab.backend import PlotHistogram

def run(threshold=0.003, nchannels=500, vmax=2000., scale=5400., 
        runtime=None, outfile=None, log=True, do_plot=True,
        input_device_index=None, inputfile=None, realtime=True,
        fitpulse=False, fit_threshold=0.95):

    if inputfile is not None:
        source=FileReplay(filename=inputfile, realtime=realtime)
    else:
        try:
            source=PyAudio(input_device_index=input_device_index)
        except:
            source=SoundCard(input_device_index=input_device_index)
    convert=Raw2Float()
    if fitpulse:
        detect=FittedPulseDetection(threshold=threshold, fit_threshold=fit_threshold)
    else:
        detect=PulseDetection(threshold=threshold)
    
    count=Count(outfile=None)
    calibrate=Scale(scale=scale)
    histogram=AggregateHistogram(nchannels=int(nchannels*vmax/1000.), vmin=0, vmax=vmax, outfile=outfile)
    
    source.plugs_into(convert)
    
    convert.plugs_into(detect)
    detect.plugs_into(count)
    detect.plugs_into(calibrate)
    calibrate.plugs_into(histogram)

    if do_plot:
        plothistogram=PlotHistogram(xmin=0,xmax=vmax, outfile=outfile, log=log)
        histogram.plugs_into(plothistogram)

    main(runtime)

def new_argument_parser():
    "Parse command line arguments"
    parser = argparse.ArgumentParser( formatter_class=argparse.ArgumentDefaultsHelpFormatter,
                                      description=__doc__)
    parser.add_argument(
        '--threshold',
        dest='threshold',
        default=0.003,
        type=float,
        help='(noise) threshold for pulse detection',
    )
    parser.add_argument(
        '--nchannels',
        dest='nchannels',
        default=100,
        type=int,
        help='number of channels per MeV (bins) in the histogram',
    )
    parser.add_argument(
        '--vmax',
        dest='vmax',
        default=2000,
        type=float,
        help='maximum energy bin for histogram',
    )
    parser.add_argument(
        '--scale',
        dest='scale',
        default=5400,
        type=float,
        help='signal energy scale (in keV)',
    )
    parser.add_argument(
        '--outfile',
        dest='outfile',
        default=None,
        help='output file root',
    )
    parser.add_argument(
        '--runtime',
        dest='runtime',
        default=None,
        type=float,
        help='runtime in seconds',
    )
    parser.add_argument(
        '--log',
        dest='log',
        action="store_true",
        help='plot logarithmic y-axis',
    )
    parser.add_argument(
        '--no-plot',
        dest='do_plot',
        action="store_false",
        help='hide plot',
    )
    parser.add_argument(
        '--input_device_index',
        dest='input_device_index',
        default=None,
        type=int,
        help='select input device',
    )
    parser.add_argument(
        '--infile',
        dest='inputfile',
        default=None,
        help='(optional) input file',
    ) 
    parser.add_argument(
        '--realtime',
        dest='realtime',
        action="store_true",
        help='file input in realtime',
    )       
    parser.add_argument(
        '--fitpulse',
        dest='fitpulse',
        action="store_true",
        help='use pulse fitting for amplitude?',
    )       
    parser.add_argument(
        '--fit-threshold',
        dest='fit_threshold',
        default=0.95,
        type=float,
        help='minimum peak to determine amplitude through fitting',
    )

    return parser.parse_args()

if __name__=="__main__":
    args = new_argument_parser()
    run(**vars(args))




#!/usr/bin/env python
"""
gammalab-monitor plots raw detector (soundcard) values. 
"""
import argparse

try:
  import matplotlib
  matplotlib.use('TkAgg')
except:
  pass

from gammalab import main
from gammalab.acquisition import PyAudio, SoundCard
from gammalab.backend import Monitor
from gammalab.transform import Raw2Float, DownSampleMaxed

def run(input_device_index=None, runtime=None):
    try:
        source=PyAudio(input_device_index=input_device_index)
    except:
        source=SoundCard(input_device_index=input_device_index)

    monitor=Monitor()
    convert=Raw2Float()
    downsample=DownSampleMaxed(factor=16)
        
    source.plugs_into(convert)
    convert.plugs_into(downsample)
    downsample.plugs_into(monitor)
  
    main(runtime)

def new_argument_parser():
    "Parse command line arguments"
    parser = argparse.ArgumentParser( formatter_class=argparse.ArgumentDefaultsHelpFormatter,
                                      description=__doc__)
    parser.add_argument(
        '--input_device_index',
        dest='input_device_index',
        default=None,
        type=int,
        help='select input device',
    )
    parser.add_argument(
        '--runtime',
        dest='runtime',
        default=None,
        type=float,
        help='runtime in seconds',
    )
    return parser.parse_args()

if __name__=="__main__":
    args = new_argument_parser()
    run(**vars(args))



from ..service import SourceService, ThreadService
from ..wire import RawWire

try:
    import soundcard
    HAS_SOUNDCARD=True
except ImportError:
    HAS_SOUNDCARD=False

class SoundCard(ThreadService, SourceService):
    def __init__(self, frames_per_buffer=2048, input_device_index=None, input_device_name="",
                 sample_rate=48000, sample_format="float32"):
        if not HAS_SOUNDCARD:
          raise Exception("soundcard module not or not correctly installed")        
        self.CHANNELS=1
        self.RATE=sample_rate
        if sample_format != "float32":
          raise Exception("SoundCard only supports float32 sample_format")
        self.FORMAT=sample_format
        self.frames_per_buffer=frames_per_buffer
        self.input_device_index=input_device_index
        self.input_device_name=input_device_name
        super(SoundCard, self).__init__()

    def _process(self):
        mic=soundcard.get_microphone(self.input_device_index or self.input_device_name)
        with mic.recorder(samplerate=self.RATE, blocksize=self.frames_per_buffer, channels=self.CHANNELS) as recorder:
            while not self.done:    
                if not self.stopped and not self.done:
                    try:
                        data=recorder.record(self.frames_per_buffer)
                    except Exception as ex:
                        self.print_message( "error: {0}".format(str(ex)))
                        self.done=True
                    if len(data)==0:
                        self.print_message("no data")
                        self.done=True
                    else:
                        data=data.astype("float32").tobytes()
    
                if (not self.stopped and 
                    not self.done and
                    data is not None):  
                    self.send_output(data)
                
            self.send_output(None)
        self.stopped=True

    def output_protocol(self, wire):
        assert isinstance(wire, RawWire)
        wire.CHANNELS=self.CHANNELS
        wire.RATE=self.RATE
        wire.FORMAT=self.FORMAT



from ..service import ReceivingService, SourceService, ThreadService
from ..wire import PulseWire

import numpy

class Interpolate(ThreadService, SourceService, ReceivingService):
    def __init__(self, calibration):
        super(Interpolate, self).__init__()
        self.input_wire=PulseWire()
        self._xp=numpy.array([c[0] for c in calibration])
        self._yp=numpy.array([c[1] for c in calibration])
        try:
            assert(numpy.all(numpy.diff(self._xp) > 0))
        except AssertionError:
            raise Exception("provide ordered calibration dataset")
        
    def output_protocol(self, wire):
        assert isinstance(wire, PulseWire)

    def process(self, data):
        timestamps=[x[0] for x in data]
        signal=[x[1] for x in data]
        calibrated=numpy.interp(signal, self._xp, self._yp)
        return zip(timestamps, calibrated)

class Scale(ThreadService, SourceService, ReceivingService):
    def __init__(self, scale):
        super(Scale, self).__init__()
        self.input_wire=PulseWire()
        self.scale=scale
        
    def output_protocol(self, wire):
        assert isinstance(wire, PulseWire)

    def process(self, data):
        timestamps=[x[0] for x in data]
        scaled=[self.scale*x[1] for x in data]
        return zip(timestamps, scaled)



from . import all_services, shared_output

import threading

class ServiceError(Exception):
    pass

class Service(object):
    def __init__(self, **kwargs):
        super(Service, self).__init__(**kwargs)
        all_services.add(self)
        self._output=shared_output
    
    def start(self):
        raise Exception("not implemented for %s"%str(self))
    def stop(self):
        raise Exception("not implemented for %s"%str(self))
    def close(self):
        raise Exception("not implemented for %s"%str(self))
    def plugs_into(self, other):
        raise ServiceError("service %s has no outputs"%str(self))
    def connect_to(self, other):
        raise ServiceError("service %s has no inputs"%str(self))
    def print_message(self, message):
        self._output.put('['+self.__class__.__name__+'] '+message)

class SourceService(Service):
    def __init__(self, **kwargs):
        super(SourceService, self).__init__(**kwargs)
        self.wires=[]

    def output_protocol(self):
        raise Exception("not implemented for %s"%str(self))

    def connect(self, wire):
        self.output_protocol(wire)
        self.wires.append(wire)

    def send_output(self,data):
        for w in self.wires:
            try:
                w.put_nowait(data)
            except:
                self.print_message("%s buffer full"%str(self))

    def plugs_into(self, other):
        try:
          other.connect_input(self)
        except ServiceError as ex:
          raise ex      
        except AssertionError as ex:
          self.print_message(str(ex))
          raise Exception("Wiring fault! Trying to connect incompatible services")
        except AttributeError as ex:
          self.print_message(str(ex))
          raise Exception("Wiring fault! Wires connected out of order?")
        
class ReceivingService(Service):
    def __init__(self, **kwargs):
        super(ReceivingService, self).__init__(**kwargs)
    def connect_input(self, service):
        service.connect(self.input_wire)
    def receive_input(self, block=True):
        try:
            return self.input_wire.get(block, 2)
        except Exception as ex:
            return None

# multiple input/outputs will be implemented by connecting to/from special attributes:
# s1.output1.plugs_into(s2)

class ThreadService(Service):
    def __init__(self, **kwargs):
        super(ThreadService, self).__init__(**kwargs)
        self.stopped=True
        self.done=False
        self.thread=threading.Thread(target=self._process)

    def start(self):
        self.stopped=False
        if not self.thread.is_alive():
            self.thread.start()

    def _process(self):
        while not self.done:
            if hasattr(self,"receive_input"):
                data=self.receive_input()
                if data is None:
                    self.done=True
            else:
                data=None

            # if stopped reveive input, but do not process, nor output

            if not self.stopped and not self.done:
                data=self.process(data)

            if hasattr(self,"send_output"):
                if (not self.stopped and 
                    not self.done and
                    data is not None):  
                    self.send_output(data)
            
        if hasattr(self,"send_output"):
            self.send_output(None)
        self.stopped=True

    def stop(self):
        self.stopped=True

    def close(self):
        self.stop()
        self.done=True
        self.thread.join()




try:
    from queue import Queue
except:
    from Queue import Queue

class Wire(Queue):
    __initialized=False
    _description="none"

    def __init__(self, **kwargs):
        Queue.__init__(self, **kwargs)
        self.protocol=dict()
        self.__initialized=True

# attributes are forwarded to the protocol (after __init__)
    def __setattr__(self, name, value):
        if self.__initialized:
            self.protocol[name] = value
        else:
            self.__dict__[name] = value

    def __getattr__(self, name):
        if name in self.protocol:
            return self.protocol[name]
        else:
            raise AttributeError("No such attribute in protocol: " + name)

    def __delattr__(self, name):
        if name in self.protocol:
            del self.protocol[name]
        else:
            raise AttributeError("No such attribute in protocol: " + name)

class RawWire(Wire):
    _description="Raw byte (string)"

class FloatWire(Queue):
    _description="numpy float array"

class PulseWire(Queue):
    _description="pulses"

class MessageWire(Queue):
    _description="output queue for terminal messages"

class HistogramWire(Queue):
    _description="dict with histogram"

